//----------------------------------------------------------------------
// vx-scheme : Scheme interpreter.
// Copyright (c) 2002,2003 and onwards Colin Smith.
//
// You may distribute under the terms of the Artistic License, 
// as specified in the LICENSE file.
//
// cell.cpp : cell creation, storage management, garbage collection.

#include "vx-scheme.h"

static const char * nomem_error = "out of memory";

Cell * Context::make ()
    {
    Cell * c = alloc (Cell::Cons);
    c->ca.p = c->cd.p = &Cell::Nil;
    return c;
    }

Cell * Context::make_int (int i)
    {
    Cell * c = alloc (Cell::Int);
    c->cd.i = i;

    return c;
    }

Cell * Context::make_char (char ch)
    {
    Cell * c = alloc (Cell::Char);
    c->cd.c = ch;
    
    return c;
    }

Cell * Context::make_real (double d)
    {
    Cell * c = alloc (Cell::Real, 2);
    ((double *) c) [1] = d;

    return c;
    }

Cell * Context::make_string (const char * s)
    {
    int len = strlen (s);
    char * s2 = (char *) malloc (len + 1);
    strcpy (s2, s);
    return make_allocated_string (s2, len);
    }

/* 
 * Use this entry point when the caller has allocated the string
 * (with malloc), and wishes to "donate" the storage to the cell
 * to avoid a gratuitous allocation and copy.  At cell finalization
 * time, the storage provided will be returned with free().
 */

Cell * Context::make_allocated_string (const char * s, int len)
    {
    Cell * c = alloc (Cell::String, 2);
    c->cd.u = s;
    c [1].ca.i = len;
    
    return c;
    }


Cell * Context::make_subr (subr_f s, const char * name)
    {
    Cell * c = alloc (Cell::Subr, 2);
    c->cd.m = s;
    c [1].cd.u = name;
    
    return c;
    }

Cell * Context::make_builtin (psymbol y)
    {
    Cell * c = alloc (Cell::Builtin);
    c->cd.y = y;
    return c;
    }

Cell * Context::make_symbol (psymbol y)
    {
    Cell * c = alloc (Cell::Symbol, 2);
    c->cd.y = y;
    c [1].ca.i = 0;  // lexical addressing
    c [1].cd.i = 0;

    return c;
    }

Cell * Context::make_boolean (bool b)
    {
    return b ? &Cell::Bool_T : &Cell::Bool_F;
    }

Cell * Context::make_vector (int n, Cell * init /* = &Unspecified */)
    {
    Cell * c = alloc (Cell::Vec, 2);
    c->cd.cv = new cellvector (n);
    c [1].ca.i = c [1].cd.i = 0; // these hold mark indices for GC
    c->flag (Cell::VREF, true);
    
    for (int ix = 0; ix < n; ++ix)
	c->cd.cv->set (ix, init);

    return c;
    }

Cell * Context::make_iport (const char * fname)
    {
    FILE * ip = fopen (fname, "r");
    if (ip)
	return make_iport (ip);

    error ("unable to open stream for reading");
    return nil;
    }

Cell * Context::make_iport (FILE * ip)
    {
    Cell * c = alloc (Cell::Iport);
    c->cd.ip = ip;

    return c;
    }

Cell * Context::make_oport (const char * fname)
    {
    FILE * ofs = fopen (fname, "w");
    if (ofs)
	return make_oport (ofs);
    
    error ("unable to open stream for writing");
    return nil;
    }

Cell * Context::make_oport (FILE * op)
    {
    Cell * c = alloc (Cell::Oport);
    c->cd.op = op;

    return c;
    }
	
Cell * Context::make (Cell * ca, Cell * cd /* = &Nil*/)
    {
    Cell * c = alloc (Cell::Cons);
    c->ca.p = ca;
    c->cd.p = cd;
    return c;
    }

Cell * Context::make_procedure (Cell * e, Cell * body, Cell * arglist)
    {
    Cell *       c  = alloc (Cell::Lambda, 2);
    cellvector * cv = new cellvector (3);

    c->cd.cv = cv;
    c->flag (Cell::VREF, true);
    cv->set (0, e);
    cv->set (1, body);
    cv->set (2, arglist);

    c [1].ca.i = c [1].cd.i = 0; // these hold mark indices for GC
    return c;
    }

Cell * Context::make_macro (Cell * e, Cell * body, Cell * arglist)
    {
    Cell * c = make_procedure (e, body, arglist);
    c->flag (Cell::MACRO, true);
    return c;
    }
    
Cell * Context::make_promise (Cell * e, Cell * body)
    {
    Cell * c = alloc (Cell::Promise, 2);
    cellvector * cv = new cellvector (1);

    // Now it may seem odd to allocate a vector of one element to
    // store the content of the promise.  But, our garbage collector
    // only knows how to traverse two kinds of entities: (1) conses,
    // consisting of a car and cdr and (2) vectors.  Since we're not a
    // cons, but contain a reference to either the procedure that will
    // compute the promise or that procedure's memoized value, we must
    // store that thing in a unit vector.

    c->cd.cv = cv;
    c->flag (Cell::VREF, true);
    c [1].ca.i = c [1].cd.i = 0; // these hold mark indices for GC
    gc_protect (c);
    cv->set (0, make_procedure (e, body, nil));
    gc_unprotect ();
    return c;
    }

Cell * Context::make_list1 (Cell * e1)
    {
    return cons (e1, nil);
    }

Cell * Context::make_list2 (Cell * e1, Cell * e2)
    {
    return cons (e1, make_list1 (e2));
    }

Cell * Context::make_list3 (Cell * e1, Cell * e2, Cell * e3)
    {
    return cons (e1, make_list2 (e2, e3));
    }

Cell * Context::make_continuation ()
    {
    Cell * c = alloc (Cell::Cont, 2);

    // Allocate a cellvector to hold the continutation (saved
    // machine stack).

    int msize = m_stack.size ();
    cellvector * cv = new cellvector (msize);
    c->flag (Cell::VREF, true);
    c->cd.cv = cv;
    c [1].ca.i = c [1].cd.i = 0; // these hold mark indices for GC

    for (int ix = 0; ix < msize; ++ix)
	cv->set (ix, m_stack [ix]);

    return c;
    }

void Context::load_continuation (Cell * cont)
    {
    cont->typecheck (Cell::Cont);

    cellvector * cv = cont->cd.cv;
    int msize = cv->size ();

    m_stack.erase ();
    for (int ix = 0; ix < msize; ++ix)
	save (cv->get (ix));
    }

Cell * Context::make_magic (void * key, magic_set_f fs, magic_get_f fg)
    {
    Cell * c = alloc (Cell::Magic, 2);
    c->cd.vp = key;
    c [1].ca.magic_set = fs;
    c [1].cd.magic_get = fg;
    return c;
    }

Cell * Cell::notcons () const
    {
    error ("expecting a Cons");
    return nil;
    }

void Cell::config ()
    {
    int bad = 0;

    printf ("size = %d, typebits = %d, typemask = %x, numtypes = %d\n",
	    sizeof (Cell), TYPEBITS, TYPEMASK, NUM_TYPES);

    // Make sure that there are enough typebits to contain
    // all the types we know about.

    if ((1 << TYPEBITS) < NUM_ATOMS)
	++bad, printf ("Not enough typebits!\n");

    // Make sure that the size of a cell has not become greater
    // than two machine pointers (car & cdr).

    if (sizeof (Cell) > 2 * sizeof (void *))
	printf ("Cell (%d) is larger than CAR+CDR!\n", sizeof (Cell));

    // Make sure that the "zero zone" (the least significant
    // bits of a pointer to a cell) is wide enough to accomodate
    // the type and GC information stored there, assuming that
    // a Cell is aligned to its own size in memory

    if (sizeof (Cell) < (1 << TAGBITS))
	++bad, printf ("Too many tag bits for cell size\n");

    if (bad)
	exit (bad);
    };

bool Cell::eq (Cell * c)
    {
    if (ca.i & ATOM || c->ca.i & ATOM) 
	{
	bool part1 = (ca.i & IGN_MASK) == (c->ca.i & IGN_MASK) 
	    && cd.i == c->cd.i;

	if (part1 && flag (DOUBLE))
	    {
	    return this [1].ca.i == c [1].ca.i
		&& this [1].cd.i == c [1].cd.i;
	    }

	return part1;
	}
    // if both are conses, they are eq iff they are the same cons.
    return this == c;
    }


bool Cell::equal (Cell * c)
    {
    Type t0 = type ();
    Type t1 = c->type ();

    if (this == &Nil && c == &Nil)
        return true;
    else if (t0 == Cons && t1 == Cons)
        return ca.p->equal (c->ca.p) && cd.p->equal (c->cd.p);
    else if (t0 == Vec && t1 == Vec)
        {
        cellvector * cv = VectorValue();
        cellvector * ocv = c->VectorValue ();
        int s = cv->size ();

        if (s != ocv->size ())
            return false;

        for (int ix = 0; ix < s; ++ix)
            if (! cv->get (ix)->equal (ocv->get (ix)))
                return false;

        return true;
        }
    else if (t0 == String && t1 == String)
        return !strcmp (StringValue (), c->StringValue ());
    else if (t0 == Real && t1 == Real)
	return RealValue () == c->RealValue ();
    else
        return eq (c);
    }

//------------------------------------------------------------------------
//
// Access/Mutate Cons Cells.  These are checked calls, in that they
// will verify that they are traversing a set of cons cells at each
// step, using "assert_cons", which throws a C++ exception if this is
// not found to be true.

void Cell::setcar (Cell * _ca)  {/*assert_cons ();*/ ca.p = _ca;       }
void Cell::setcdr (Cell * _cd)  {/*assert_cons ();*/ cd.p = _cd;       }
//Cell * Cell::car () const          {return ca.i & ATOM ? notcons () : ca.p; }
//Cell * Cell::cdr () const          {return cd.i & ATOM ? notcons () : cd.p; }
Cell * Cell::caar ()            {return car ()->car ();            }
Cell * Cell::cadr ()            {return cdr ()->car ();            }
Cell * Cell::cdar ()            {return car ()->cdr ();            }
Cell * Cell::cddr ()            {return cdr ()->cdr ();            }
Cell * Cell::caaar ()           {return caar ()->car ();           }
Cell * Cell::caadr ()           {return cadr ()->car ();           }
Cell * Cell::cadar ()           {return cdar ()->car ();           }
Cell * Cell::caddr ()           {return cddr ()->car ();           }
Cell * Cell::cdaar ()           {return caar ()->cdr ();           }
Cell * Cell::cdadr ()           {return cadr ()->cdr ();           }
Cell * Cell::cddar ()           {return cdar ()->cdr ();           }
Cell * Cell::cdddr ()           {return cddr ()->cdr ();           }
Cell * Cell::caaaar ()          {return caaar ()->car ();          }
Cell * Cell::caaadr ()          {return caadr ()->car ();          }
Cell * Cell::caadar ()          {return cadar ()->car ();          }
Cell * Cell::caaddr ()          {return caddr ()->car ();          }
Cell * Cell::cadaar ()          {return cdaar ()->car ();          }
Cell * Cell::cadadr ()          {return cdadr ()->car ();          }
Cell * Cell::caddar ()          {return cddar ()->car ();          }
Cell * Cell::cadddr ()          {return cdddr ()->car ();          }
Cell * Cell::cdaaar ()          {return caaar ()->cdr ();          }
Cell * Cell::cdaadr ()          {return caadr ()->cdr ();          }
Cell * Cell::cdadar ()          {return cadar ()->cdr ();          }
Cell * Cell::cdaddr ()          {return caddr ()->cdr ();          }
Cell * Cell::cddaar ()          {return cdaar ()->cdr ();          }
Cell * Cell::cddadr ()          {return cdadr ()->cdr ();          }
Cell * Cell::cdddar ()          {return cddar ()->cdr ();          }
Cell * Cell::cddddr ()          {return cdddr ()->cdr ();          }

psymbol Cell::SymbolValue () const
    {
    typecheck (Symbol);
    return cd.y;
    }

void Cell::stats ()
    {
    if (!isatty (0))
	return;

    for (int ix = 0; ix < NUM_TYPES; ++ix)
        printf ("%s %d ", typeName [ix], typeCount [ix]);

    printf ("\n");
    }

//======================================================================
// 
//              Value Extractors
//
//======================================================================

int Cell::IntValue () const
    {
    typecheck (Int); return cd.i;
    }

char Cell::CharValue () const
    {
    typecheck (Char);
    return cd.c;
    }

subr_f Cell::SubrValue () const
    {
    typecheck (Subr);
    return cd.m;
    }

char * Cell::StringValue () const
    {
    typecheck (String);
    return cd.s;
    }

int Cell::StringLength () const 
    {
    typecheck (String);return this[1].ca.i;
    }

FILE * Cell::IportValue () const
    {
    typecheck (Iport); return cd.ip;
    }

FILE * Cell::OportValue () const
    {
    typecheck (Oport); return cd.op;
    }

void * Cell::ContValue () const
    {
    typecheck (Cont); return cd.j;
    }

cellvector * Cell::VectorValue () const
    {
    typecheck(Vec); return cd.cv;
    }

Cell * Cell::PromiseValue () const
    {
    typecheck (Promise);
    return cd.cv->get (0);
    }

psymbol Cell::BuiltinValue () const
    {
    typecheck (Builtin); return cd.y;
    }

Cell::Procedure Cell::LambdaValue () const
    {
    typecheck (Lambda);
    return Procedure (cd.cv->get (0), cd.cv->get (1), cd.cv->get (2));
    }

double Cell::RealValue () const
    {
    typecheck (Real);
    return reinterpret_cast <const double*> (this) [1];
    }

const char * Cell::name () const
    {
    return typeName [type ()];
    }

void Cell::typefail (Type t1, Type t2) const
    {
    static char buf [128]; // XXX not reentrant, and fixed buffer dangerous
    sprintf (buf, "type check failure: wanted %s, got %s",
	     typeName [t2], typeName [t1]); /* XXX sprintf into fixed buf */
    
    OS::exception (buf);
    }

void Cell::dump (FILE * out)
    {
    Type t = type ();
    fprintf (out, "[%p ", this);

    if (ca.i & MARK) fputs ("mark ", out);
    if (ca.i & FREE) fputs ("free ", out);
    if (ca.i & ATOM) 
	{
	printf ("atom %04x ", ca.i);
	if (ca.i & FORCED) fputs ("forced ", out);
	if (ca.i & QUICK)  fputs ("quick ", out);
	if (ca.i & MACRO)  fputs ("macro ", out);
	if (ca.i & DOUBLE) fputs ("double ", out);
	if (ca.i & VREF)   fputs ("vref ", out);
	}
    
    fputs (typeName [t], out);

    switch (t)
	{
	case Cons:
	    fputs (" ", out);
	    if (ca.p == nil)
	        fputs ("nil", out);
	    else
	        fprintf (out, "%p", ca.p);
	    fputs (" ", out);
	    if (cd.p == nil)
	        fputs ("nil", out);
	    else
	        fprintf (out, "%p", cd.p);
	    break;

	case Int:    fprintf (out, " %d", cd.i); break;
	case Real:   fprintf (out, " %g", RealValue ()); break;
	case Unique: fprintf (out, " %s", cd.u); break;
	default: break;
	}

    fputc (']', out);
    }
    

//======================================================================
// 
//              Cell Vectors
//
//======================================================================

cellvector::cellvector (int _size /* = 0 */)
    {
    alloc = (_size == 0 ? 10 : _size);

    v = (Cell **) malloc (alloc * sizeof (Cell *));
    if (!v)
	error (nomem_error);

    for (int ix = 0; ix < alloc; ++ix)
	v [ix] = nil;

    sz = _size;
    }

Cell *& cellvector::operator [] (int ix)
    {
    if (ix < 0 || ix >= sz)
	vref_error ();

    return v [ix];
    }

Cell * cellvector::get (int ix)

    {
    if (ix < 0 || ix >= sz)
	vref_error ();

    return v [ix];
    }

void cellvector::set
    (
    int          ix,
    Cell *       c
    )

    {
    if (ix < 0 || ix >= sz)
	vref_error ();

    v [ix] = c;
    }

Cell * cellvector::top ()
    {
    if (sz <= 0)
	vref_error ();

    return v [sz - 1];
    }

void cellvector::expand ()
    {
    // Must expand vector: double size.
    int     new_alloc = 2 * alloc;
    Cell ** v2 = (Cell **) malloc (new_alloc * sizeof (Cell *));

    if (!v2)
	error (nomem_error);

    memcpy (v2, v, alloc * sizeof (Cell *));
    free (v);
    v = v2;
    alloc = new_alloc;
    }

//void cellvector::push (Cell * c) 
//    {
//    if (sz == alloc)
//        expand (); 
//    v [sz++] = c; 
//    }

//Cell * cellvector::pop () 
//    {
//    if (sz <= 0) 
//        error (vref_error);  
//    return v [--sz]; 
//    }

void cellvector::vref_error ()
    {
    error (error_text);
    }

void cellvector::discard (int n /* = 1 */)
    {
    if (n < 0 || n > sz)
	vref_error ();
    sz -= n;
    }

void cellvector::erase ()
    {
    sz = 0;
    }

cellvector::~cellvector ()

    {
    free (v);
    sz = 0;
    alloc = 0;
    v = 0;
    }


//======================================================================
// 
//              Memory Allocation and Garbage Collection
//
//======================================================================

class Slab 
    {
    public:
	
    Cell *      alloc (int size)
        {
	if (next + size > end)
	    return 0;

	Cell * r = next;
	next += size;
	return r;
	}

    int remaining ()
	{
	return end - next;
	}

    void reset ()
	{
	next = start;
	}

    void sweep (Context *);

    Slab (Context * ctx)
        {
	// We avoid the temptation to call new Cell [slabsize], 
	// since that would invoke the constructor on each cell,
	// which we don't need (alloc will take care of preparing
	// cells for use).  
	//
	// It is essential that Cells be 8-aligned to preserve
	// three bits for type and GC information.  If new has
	// stiffed us with 4-aligned memory, we "burn" 4 bytes
	// of it.

	int storage_size = slabsize * sizeof (Cell) + 4;
	storage = (char *) malloc (storage_size);

	if (!storage)
	    error ("out of memory");

	// Supposedly the ANSI library guarantees that storage
	// is 4-aligned!

	if (((int) storage) & 3)
	    abort ();

	// But if it's not 8-aligned we can fix that using the 
	// extra 4 bytes we allocated.

	if (((int) storage) & 7)
	    start = reinterpret_cast <Cell *> (storage + 4);
	else
	    start = reinterpret_cast <Cell *> (storage);

	memset (storage, 0, storage_size);

	ctx->cellsTotal += slabsize;
	end = start + slabsize;
	reset ();
	}
	
    ~Slab ()
	{
	free (storage);
	}

    static int slabsize;

    private:

    Cell *	start;
    Cell *      end;
    Cell *      next;
    char *      storage;
    };

int Slab::slabsize = 10000;

Cell * Context::alloc (Cell::Type t, int s /* = 1 */)
    {
    Cell * a;

    mem.last_alloc_gc = false;
    // Select a cell from the free list if one is available.

TOP:
    if ((s == 1) && (a = mem.free1))
	{
	cellsAlloc += s;
	mem.free1 = a->cd.p;
	a->ca.i = a->cd.i = 0;
	a->set_type (t);
	--mem.c_free1;
	return a;
	}

    if ((s == 2) && (a = mem.free2))
	{
	cellsAlloc += s;
	mem.free2 = a->cd.p;
	a->ca.i = Cell::DOUBLE;
	a->cd.i = a [1].ca.i = a [1].cd.i = 0;
	a->set_type (t);
	--mem.c_free2;
	return a;
	}

    // IF there aren't any slabs in the active pool, 
    // we must never have allocated any slabs at all
    // yet, so allocate the first one.

    if (mem.active.size () == 0)
	{
	// Configurable slabsize

	char * c;
	if ((c = getenv ("SLABSIZE")) != NULL)
	    Slab::slabsize = atoi (c);

	mem.active.push ((Cell *) new Slab (this));
	mem.free1 = mem.free2 = 0;
	mem.low_water = false;
	mem.no_inline_gc = OS::flag (DEBUG_NO_INLINE_GC);
	}

    // Check the "top" slab to see if there's any room 
    // left in it.

    if ((a = mem.current ()->alloc (s)))
	{
	cellsAlloc += s;
	a->ca.i = (s == 2) ? Cell::DOUBLE : 0;
	a->cd.i = 0;
	a->set_type (t);
	return a;
	}

    // There wasn't any room in the top slab.  We can try 
    // to GC.  If we do, and still 80% of the allocated 
    // memory is occupied, we set a flag admitting that 
    // the last GC was "unproductive", and next time 'round
    // we'll allocate a new slab.

    if (mem.no_inline_gc || mem.last_alloc_gc || mem.low_water)
	{
	mem.active.push ((Cell *) new Slab (this));  // trip to the well
	mem.low_water = false;                       // low_water is a one-shot
	}
    else
	{
	mem.last_alloc_gc = true;
	gc ();
	}

    goto TOP;
    }

//----------------------------------------------------------------------
// GARBAGE COLLECTION
//


inline Cell * Cell::untagged (Cell * c)
    {
    return reinterpret_cast <Cell *>
	  (reinterpret_cast <int> (c) & ~Cell::TAGMASK);
    }

void Cell::gc_set_car (Cell * src)
    {
    unsigned int tagbits = ca.i & TAGMASK;
    ca.p = src;
    ca.i |= tagbits;
    }

void Cell::gc_set_cdr (Cell * src)
    {
    unsigned int tagbits = cd.i & TAGMASK;
    cd.p = src;
    cd.i |= tagbits;
    }

    
//----------------------------------------------------------------------
// Marking for Garbage Collection
//
// This implementation is Knuth's Algorithm 2.3.5E (TAoCP 3ed. vol I
// p. 418) We follow Knuth's presentation carefully (using the same
// variable names and statement labels).  Like the evaluator, this 
// code has to take some care to avoid recursion: we want to be able
// to perform a GC mark wihtout allocating any additional space (not 
// even C stack space).  That accounts for some of the complexity in 
// this routine.  The other part is that, due to vectors, we have to 
// support n-way marking instead of just 2-way marking.

void Context::mark (Cell * P)
    {
    if (P == nil || P == 0 || P->ca.i & Cell::MARK)
	return;

    // In Knuth's presentation, a NODE contains two pointers
    // (which he calls ALINK and BLINK, we car and cdr), and 
    // MARK and ATOM fields.  In his layout, the MARK and 
    // ATOM fields can be manipulated easily without changing
    // ALINK and BLINK, but in our case we store MARK and ATOM
    // in the lower three bits of ALINK.  We must therefore
    // be cautious when transcribing the algorithm to avoid, 
    // e.g., clearing MARK and ATOM when copying a `car' pointer.
    // We use "gc_set_car" and "gc_set_cdr" for this purpose.

//E1:
    Cell * T = nil;
    Cell * Q = nil;
    bool traceall = OS::flag (TRACE_GC_ALL);

E2: P->ca.i |= Cell::MARK;   
    if (traceall) { printf ("m "); P->dump (stdout); putchar ('\n'); }

    // -- EXTENSION to Knuth's Step E2 
    //
    // If the cell is a cons, Knuth's algorithm will take care of
    // marking the things referenced as a result quite handily.  But
    // there are some atoms that can hold references too.  Knuth's
    // algorithm works for binary trees, but to deal with vectors et
    // al. we need to make it work for n-way trees.
    //
    // When an atom can hold references to other cells, we organize
    // these into a cellvector.  In this way, we can treat all of them
    // the same way.
    //
    // Whereas Knuth uses an atom bit to tell which side of a cons
    // (car or cdr) he has stashed the pointer back to the
    // as-yet-unmarked cells, when we traverse a vector we use an
    // auxiliary integer field to tell us how many vector slots we
    // have marked so far.

    if (P->flag (Cell::VREF))
	{
	// Getting "here" in the code means that we're seeing the
	// vector of additional cell references for the first time
	// (otherwise the mark bit will already be set).  Our job is
	// to kick off the iteration by stashing the back-link and
	// starting the mark counter.  The rest of the iteration will
	// be handled in the "up" step below.

	if (P->cd.cv->size () > 0)
	    {
	    P [1].cd.p = T; // stash uplink
	    P [1].ca.i = 0; // initialize counter
	    T = P;
	    }
	}

    if (P->ca.i & Cell::ATOM)                   // E3 
        goto E6;

    Q = Cell::untagged (P->ca.p);               // E4
    if (Q != nil && !(Q->ca.i & Cell::MARK))
	{
	P->cd.i |= Cell::ATOM;
	P->gc_set_car (T);
	T = P;
	P = Q;
	goto E2;
	}
E5: Q = Cell::untagged (P->cd.p);
    if (Q != nil && !(Q->ca.i & Cell::MARK))
	{
	P->gc_set_cdr (T);
	T = P;
	P = Q;

	goto E2;
	}
E6: if (T == nil)
	return;

    Q = T;
    
    if (Q->flag (Cell::VREF))
	{
	// We are popping a vector cell from the GC stack.
	// If there are more cells to mark within it, keep
	// going.
	
    next_element:
	
	int i = (Q [1].ca.i)++;

	if (i >= Q->cd.cv->size ())   // all done?
	    {
	    Q [1].ca.i = 0;                        // reset for next time
	    T = Q [1].cd.p;                        // pop stack
	    Q [1].cd.p = 0;
            P = Q;
	    goto E6;
	    }
	else                                        // resume iteration
	    {
	    P = Q->cd.cv->get (i);                  // with next element

	    // One wrinkle: captured continuations are implemented
	    // as vectors, and like the machine stack, these vectors
	    // can contain integer VM codes as well as cell pointers.
	    // These latter are marked with the ATOM flag.

	    if (reinterpret_cast <int> (P) & Cell::ATOM)
		goto next_element;
	    
	    // Otherwise we mark, if not marked already.
	    if (P->ca.i & Cell::MARK)
		goto next_element;

	    P = Cell::untagged (P);
	    if (P == nil)
		goto next_element;

	    goto E2;
	    }
	}

    if (Q->cd.i & Cell::ATOM)
	{
	Q->cd.i &= ~Cell::ATOM;
	T = Cell::untagged (Q->ca.p);
	Q->gc_set_car (P);
	P = Q;
	goto E5;
	}
    else
	{
	T = Cell::untagged (Q->cd.p);
	Q->gc_set_cdr (P); 
	P = Q;
	goto E6;
	}
    }

void Slab::sweep (Context * ctx)
    {
    int csize = 1;
    bool traceall = OS::flag (TRACE_GC_ALL);

    for (Cell * p = start; p < next; p += csize)
	{
	csize = p->flag (Cell::DOUBLE) ? 2 : 1;
	int word = p->ca.i;

	if (word & Cell::MARK)
	    {
	    p->ca.i &= ~Cell::MARK;
	    }
	else if (! (word & Cell::FREE))
	    {
	    // FINALIZATION
	    //

	    if (traceall) { printf ("s "); p->dump (stdout); putchar ('\n'); }
	    Cell::Type t = p->type ();

	    switch (t)
		{
		case Cell::Cont:
		case Cell::Promise:
		case Cell::Lambda:
		case Cell::Vec:     // Free the vector of cell pointers.
		    delete p->cd.cv;
		    p->cd.cv = 0;
		    break;

		case Cell::String:  // Strings hold malloc'd buffers.
		    free (p->cd.s);
		    break;

		case Cell::Iport:   // Ports hold streams
		    fclose (p->cd.ip);
		    break;
		    
		case Cell::Oport:   // Ports hold streams
		    fclose (p->cd.op);
		    break;

		default:            // Ordinarily cells hold no other storeage.
		    ;
		}


	    ctx->cellsAlloc -= csize;
	    p->ca.i = Cell::FREE;

	    if (csize == 2)
		{
		// the sweeper needs to know the size of free cells, so 
		// we make sure this one is still marked double.
		p->ca.i |= Cell::ATOM | Cell::DOUBLE;
		p->cd.p = ctx->mem.free2;
		ctx->mem.free2 = p;
		++ctx->mem.c_free2;
		}
	    else
		{
		p->cd.p = ctx->mem.free1;
		ctx->mem.free1 = p;
		++ctx->mem.c_free1;
		}
	    }
	}
    }
  
void Context::gc ()
    {	
    bool gc_verbose = OS::flag (TRACE_GC);
    Cell * p;

    if (!ok_to_gc) 
	{
	fprintf (stderr, "initial memory budget insufficient to set up VM\n");
	exit (1);
	}
    if (gc_verbose) 
        printf ("; start gc: %d/%d\n", cellsAlloc, cellsTotal);

    // 
    // MARK PHASE
    //
    // We have to mark everything reachable from the "register machine"
    // registers.

    mark (r_env);
    mark (r_argl.car ());
    mark (r_argl.cdr ());
    mark (r_varl.car ());
    mark (r_varl.cdr ());
    mark (r_proc);
    mark (r_exp);
    mark (r_unev);
    mark (r_val);
    mark (r_tmp);
    mark (r_elt);
    mark (r_nu);

    // Mark everything reachable from the machine stack.  Watch out
    // for integers hiding in the machine stack, though!  They are 
    // marked with the ATOM flag.

    for (int ix = 0; ix < m_stack.size (); ++ix)
	if ((reinterpret_cast <int> ((p = m_stack [ix])) & Cell::ATOM) == 0)
	    mark (p);

    // Mark the I/O ports referenced in this environment stack.

    for (int ix = 0; ix < istack.size (); ++ix)
	mark (istack [ix]);
    for (int ix = 0; ix < ostack.size (); ++ix)
	mark (ostack [ix]);

    // Mark the things that "C" implementations of Scheme functions
    // have requested protection for.

    for (int ix = 0; ix < r_gcp.size (); ++ix)
	mark (r_gcp [ix]);

    // 
    // SWEEP PHASE 
    //

    for (int ix = 0; ix < mem.active.size (); ++ix)
	((Slab *) mem.active [ix])->sweep (this);

    // If this mark/sweep phase managed to reduce the cell utilization
    // to <= 80% of the allocated cells, we consider that success.  On
    // the other hand, if the GC produced less than 20% free cells, we
    // set a flag which will provoke the allocation of a new slab at
    // the next allocation failure.  In this way we hope to avoid
    // "grinding away" at the last few cells in a slab.

    if ((double) cellsAlloc / cellsTotal > 0.8)
	mem.low_water = true;

    if (gc_verbose)
        printf (";   end gc: %d/%d %s\n", cellsAlloc, cellsTotal,
		mem.low_water ? " low" : " ok");
    }

void Context::gc_if_needed ()
    {
    if (cellsAlloc >= cellsTotal / 4 * 3)
	gc ();
    }

void Context::print_mem_stats (FILE * out)
    {
    fprintf (out, "; mem %d/%d\n", cellsAlloc, cellsTotal);
    }

