//----------------------------------------------------------------------
// vx-scheme : Scheme interpreter.
// Copyright (c) 2002,2003 and onwards Colin Smith.
//
// You may distribute under the terms of the Artistic License, 
// as specified in the LICENSE file.
//
// vx-scheme.h : class definitions

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <limits.h>
#ifndef WIN32
#include <unistd.h>
#else
#define isatty(fd) 0
#endif
#if __GNUG__ >= 3
using namespace std; // so sue me
#endif

class OS;
class Cell;
class Context;    

// OS abstraction layer

class OS
    {
    public:

    static void         get_time (int & sec, int & usec); // get timestamp
    static bool         interactive (int fd);             // terminal input?
    // supply value for undef symbol
    static Cell *       undef (Context *, const char *); 
    // report exception and restart
    static void         exception (const char *); 
    // manage debug flags
    static unsigned int flags ();      
    static bool         flag (int bit)
	{ return (flags () & bit) != 0; } 
    };

typedef Cell * (* subr_f)      (Context * ctx, Cell * arglist);
typedef void   (* magic_set_f) (Context *, void * key, Cell * rhs);
typedef Cell * (* magic_get_f) (Context *, void * key);
extern  Cell * nil;
extern  Cell * unspecified;
extern  Cell * unimplemented;

void error (const char *, const char * = 0);

// FOR_EACH is a macro that can be used to traverse a standard Scheme
// list.  The variable `var' is bound for the duration of the traversal
// to each node in the list.  

#define FOR_EACH(var,list) \
    for (Cell * var = list; var != nil; var = var->cdr ())

// The symbol table is implemented as an AVL tree of these nodes.
// There's no repetition, so the address of one of these nodes can
// serve as a unique hashcode for a symbol for equality-testing
// purposes.  There's one call, intern(), for introducing a new
// string to the collection.
//
// The Scheme standard, however, introduces one complication: the 
// requirement that symbols be stored in a "standard case."  This 
// is in conflict with our desire to have case-sensitive symbol
// matching (for integration with underlying symbol tables).  (Scheme
// also provides the primitive string->symbol, which can be used to 
// create symbols outside of standard case, but the REPL is not 
// expected to use this.)
// 
// In the end I decided to spend some extra memory to achieve standard
// compliance and VxWorks symbol table integration at the same time.
// We choose to consider lower-case symbols as "canonical".  (The standard
// says we must choose upper or lower case, but not which one).  In the event
// that a symbol arrives in mixed case, we store it both ways: canonically 
// (that is, with lowered case) for Scheme symbol lookup, and unmolested
// so that, when we try the VxWorks symbol table after all else has failed,
// we can respect the case of the sybmol as written.

typedef struct          _symbol
    {
    struct _symbol *    llink;              /* Left binary tree link */
    struct _symbol *    rlink;              /* Right binary tree link */
    const char *        key;                /* Search key (symbol name */
    const char *        truename;
    int                 b;                  /* Balance factor */
    }                   symbol, *psymbol;

psymbol                 intern (const char * name);
psymbol                 intern_stet (const char * name);
Cell *                  vector_from_list (Context * ctx, Cell *);
Cell *                  vector_to_list (Context * ctx, Cell *);


class cellvector
    {
    public:

    cellvector (int size = 0);
    ~cellvector ();
    
    Cell *      get (int);
    void        set (int, Cell *);
    Cell *&     operator [] (int);
    Cell *      top ();
    void        push (Cell * c) 
	{ if (sz == alloc) expand (); v [sz++] = c; }
    Cell *      pop ()
	{ if (sz <= 0) vref_error (); return v [--sz]; }
	
    int         size () {return sz;}
    void        discard (int n = 1);
    void        erase ();

    private:

    void        expand ();
    void        vref_error ();
    int         sz;
    int         alloc;
    Cell **     v;
    static const char * const error_text = "vector reference out of bounds";
    };

// ------------------------------------------------------------------------
// class sio: input/output behavior we expect from strings or streams.
//            interface class.

class sio 
    {
    public:
    virtual int get ()     = 0;
    virtual int peek ()    = 0;
    virtual bool eof ()    = 0;
    virtual void unget ()  = 0;
    virtual void ignore () = 0;
    };

// ------------------------------------------------------------------------
// class file_sio: This wraps a FILE* into an object that answers to 
//                 the above interface.

class file_sio : public sio 
    {
    public:

    file_sio (FILE * _fp) : fp (_fp), lastch (-1) {};

    virtual int  get    () { return lastch = fgetc (fp); } 
    virtual int  peek   () { int c = get (); ungetc (c, fp); return c; } 
    virtual bool eof    () { return feof (fp) != 0; }
    virtual void unget  () { ungetc (lastch, fp); }
    virtual void ignore () { get (); } 

    private:

    FILE *          fp;
    int             lastch;
    };

// ------------------------------------------------------------------------
// An "sstring" is a simple extensible string.  It reallocates storage
// as necessary to support arbitrary growth.  It is a poor cousin to
// STL's string, but with considerably less code-bloat since there's no
// template expansion or nontrivial inlining.
// 
// In order to avoid involving the strstream class, we also extend 
// our sstream with a small amount of I/O semantics.  This allows 
// sstrings to be passed to the lexical analyzer.


class sstring : public sio
    {
    public:

    sstring ();
    virtual ~sstring ();

    char * str ()
	{ return base; } 
    char & operator [] (int ix)
	{ return base [ix]; } 
    
    void append (const char *);
    void append (const char *, int len);
    void append (const char);
    int length ()
	{ return sz; }
    void claim ();      // claim dynamic storage
    bool operator == (const char * s)
	{ return !strcmp (base, s); }

    // I/O behavior

    int get ();
    int peek ();
    bool eof ();
    void unget ();
    void ignore ();

    private:

    static const int    stat_size = 32;
    char                c [stat_size];

    int                 sz;
    int                 alloc;
    char *              base;
    char *              end;
    char *              pos;   // I/O read position
    bool                claimed;
    };

//----------------------------------------------------------------------
// class Cell
//
// The Cell is the heart of the Scheme implementation.  It is the 
// universal container for all Scheme data types and also the central
// structure supporting Scheme's garbage-collected memory model. 
// The economy of a Cell's realization is the single most significant
// factor influencing the speed and space efficiency of a Scheme
// system (with the possible exception of compilation, beyind the 
// scope of this header file).  
//
// We consider it imperative that an ordinary cell be no larger than 
// two machine pointers (car and cdr); if a data object requires 
// more storage than this, we allocate extension words.
// 
// For our implementation, we expect that a machine pointer is at
// least four bytes, so that two of these (car,cdr) will occupy 
// eight bytes.  In consequence, we may therefore insist that the 
// storage for cells be 8-byte aligned, which gives us three bits at
// the least-significant end of a cell pointer to use as type-tagging
// information.  We use two of these bits: one is the ATOM flag, which 
// indicates that a given cell is not to be interpreted as a (car,cdr)
// pair but as a container for another type of Scheme data, and the 
// DOUBLE flag, which indicates this is a "double decker" cell.

class Cell
    {
    friend class Context;
    friend class Slab;
    
public:
    
    typedef Cell * ptr;
    
    void display (FILE *);
    void write (FILE *) const;

    bool eq (Cell * c);
    bool eqv (Cell * c)
        {
        return eq (c);
        }

    bool equal (Cell * c);
    bool is_symbol (psymbol s)
	{
	return type () == Cell::Symbol && SymbolValue () == s;
	}

    struct Procedure
        {
        Procedure (Cell * _envt, Cell * _body, Cell * _arglist)
            : body (_body),
              arglist (_arglist),
              envt (_envt)
            {}

	Procedure ()
	    : body (nil),
	      arglist (nil),
	      envt (nil)
	    {}

        Cell *        body;
        Cell *        arglist;
	Cell *        envt;
        };

    // Certain cells we have heard of

    static Cell	Nil;
    static Cell	Unspecified;
    static Cell Eof_Object;
    static Cell Bool_T;
    static Cell Bool_F;
    static Cell Apply;
    static Cell Error;
    static Cell Unimplemented;

    // Access/Mutate Cons Cells.  These are checked calls, in 
    // that they will verify that they are traversing a set of 
    // cons cells at each step, using "assert_cons", which 
    // throws a C++ exception if this is not found to be true.

    void setcar (Cell * car);
    void setcdr (Cell * cdr);
    Cell * car () const { return ca.i & ATOM ? notcons () : ca.p; }  
    Cell * cdr () const { return ca.i & ATOM ? notcons () : cd.p; } 
    Cell * caar ();
    Cell * cadr ();          
    Cell * cdar ();          
    Cell * cddr ();          
    Cell * caaar ();         
    Cell * caadr ();         
    Cell * cadar ();         
    Cell * caddr ();         
    Cell * cdaar ();         
    Cell * cdadr ();         
    Cell * cddar ();         
    Cell * cdddr ();         
    Cell * caaaar ();        
    Cell * caaadr ();        
    Cell * caadar ();        
    Cell * caaddr ();        
    Cell * cadaar ();        
    Cell * cadadr ();        
    Cell * caddar ();        
    Cell * cadddr ();        
    Cell * cdaaar ();        
    Cell * cdaadr ();        
    Cell * cdadar ();        
    Cell * cdaddr ();        
    Cell * cddaar ();        
    Cell * cddadr ();        
    Cell * cdddar ();        
    Cell * cddddr ();        

    // Value extractors

    int          IntValue ()     const;
    char         CharValue ()    const;
    subr_f       SubrValue ()    const;
    char *       StringValue ()  const;
    int          StringLength () const;                 
    FILE *       IportValue ()   const;
    FILE *       OportValue ()   const;
    void *       ContValue ()    const;
    cellvector * VectorValue ()  const;
    Cell *       PromiseValue () const;
    psymbol      SymbolValue ()  const;
    psymbol      BuiltinValue () const;
    Procedure    LambdaValue ()  const;
    double       RealValue ()    const;
    const char * name ()         const;

    static void         real_to_string (double, char *, int);
    
    // In scheme, the only two values of type `boolean' are #t and 
    // #f.  However, from the point of view of truth valuation, 
    // anything other than #f is considered `true'.  We follow the
    // Scheme standard strictly, and so do not consider nil to have
    // a false connotation as it would in other dialects of Lisp.

    bool isBoolean ()        {return this == &Bool_T || this == &Bool_F;}
    bool istrue ()           {return this != &Bool_F;                   }

    static Cell * untagged (Cell *);
    
    // Utilities
    int length ()
        {
        int i = 0;

	FOR_EACH (p, this)
	    ++i;
	
        return i;
        }

    class List
	{
	public:
	
	List () : h (&Nil), t (&Nil)
	    {}

	void append (Cell * c)
	    {
	    if (t == &Nil)
		h = t = c;
	    else
		{
		t->setcdr (c);
		t = c;
		}
	    }

	Cell * head () { return h; }
	Cell * tail () { return t; } 

	private:

	Cell * h;	// head
	Cell * t;	// tail
	};
	
    void list_append (Cell *& head, Cell *& tail)
	{
	if (tail == &Nil)
	    {
	    head = tail = this;
	    }
	else
	    {
	    tail->setcdr (this);
	    tail = this;
	    }
	}
    
    static void stats ();
    static void config ();

    enum Type
        {
	//------------
        Int       = 0,  // The Atoms...
        Symbol    = 1,
        Unique    = 2,
        String    = 3,
        Real      = 4,
        Subr      = 5,
        Lambda    = 6,
	Vec       = 7,
	Char      = 8,
        Iport     = 9,
        Oport     = 10,
        Promise   = 11,
	Cont      = 12,
	Builtin   = 13,
        Magic     = 14,

	NUM_ATOMS = 15,
	//------------
	Cons      = NUM_ATOMS,  // A cell.
	NUM_TYPES = Cons + 1
	//------------
        };

    Type type () const
	{
	return (Type) ((ca.i & ATOM)
	    ? ((ca.i >> TAGBITS) & TYPEMASK)
	    : Cons);
	}

    bool macro () const
	{
	return flag (MACRO);
	}

    private:

    void    gc_set_car (Cell *);
    void    gc_set_cdr (Cell *);
    Cell *  notcons () const;

    Cell ()
        {
	ca.p = cd.p = &Nil;
        }
    
    Cell (const char * unique_name)
        {
        ca.i = 0;
        set_type (Unique);
        cd.u = unique_name;
        }

    void typecheck (Type t) const
        {
	if (type () != t)
            typefail (type (), t);
	}

    void typefail (Type t1, Type t2) const;

    static const int    ATOM      = 0x1;
    static const int	MARK	  = 0x2;
    static const int    FREE      = 0x4;
    static const int    TAGBITS   = 3;
    static const int    TYPEBITS  = 5;
    static const int    TYPEMASK  = (1 << TYPEBITS) - 1;
    static const int    TAGMASK   = (1 << TAGBITS) - 1;
    // Make sure flag bits are disjoint from TYPE and TAG bits.
    static const int    FLAGBASE  = 1 << (TYPEBITS + TAGBITS);
    static const int    FORCED    = FLAGBASE;
    static const int    QUICK     = FLAGBASE << 1;
    static const int    MACRO     = FLAGBASE << 2;
    static const int    DOUBLE    = FLAGBASE << 3;
    static const int    VREF      = FLAGBASE << 4;

    // The set of bits which should be ignored when 
    // comparing two cells in the sense of "eq?".

    static const int    IGNORE    = QUICK;
    static const int    IGN_MASK  = ~IGNORE;

    static const char * typeName  [NUM_TYPES];
    static int		typeCount [NUM_TYPES];

    void flag (unsigned int f, bool b)
        {
        if (b)
            ca.i |= f;
        else
            ca.i &= ~f;
        }

    void dump (FILE *);
    bool flag (unsigned int f) const
        {
	// only atoms can have flags.  All requested bits must be set
        return (ca.i & (f | ATOM)) == (f | ATOM);
        }

    void set_type (Type t)
	{
	if (t != Cons)
	    ca.i |= (t << TAGBITS) | ATOM;
	++typeCount [t];
	}

    // The actual data for an Atom/Cell is here.

    union _car
        {
        unsigned int        i;
        Cell *              p;
	void *              vp;
	magic_set_f         magic_set;
        }                   ca;
    union _cdr
        {
        unsigned int        i;
        Cell *              p;
        char *              s;
        const char *        u;
        subr_f              m;
	psymbol             y;
        Cell *              e;
	cellvector *	    cv;
        FILE *              ip;
        FILE *              op;
	char		    c;
	void *              vp;
	magic_get_f         magic_get;
        void *              j;
        }                   cd;
    };

//----------------------------------------------------------------------
// class Environment
//
// At the simplest level, an Environment is a mapping from symbols
// to values.  Symbols are the hash codes maintained by the SymbolTable
// class, and the value of any symbol is simply a pointer to a Scheme
// cell.  To implement this simple data structure, we use an STL vector
// of <symbol, Cell*> pairs.  This choice of data structure is guided
// by some particularities of evaluation in Scheme (discussed below).
//
// Environments are created by binding constructs (like let and lambda),
// and a new environment is always linked to the environment in force
// when it was created (this is called the "enclosing environment").
// The enclosure chain always terminates at the global environment, which 
// is where the symbols representing the language's standard features
// are bound.  
//
// In Scheme, all variables are "lexically bound."  This means that
// when a variable is mentioned in source code, one can determine the
// binding for that variable at "compile time" by looking through the
// stack of bindings crated by special forms capable of creating such
// bindings (e.g., lambda, let, et al.).  The innermost matching binding
// found represents the storage for the value of the variable, and
// this can never change.
//
// This binding model creates the possibility of lexcial addressing, a
// system in which a variable reference can be replaced by the "index"
// of the storage in terms of the number of enclosing environments
// that must be traversed together with the index of the target
// variable within that environment.  This represents an extremely
// efficient shortcut for variable value lookup.  This is why we
// choose the vector data structure rather than an STL map: while a
// vector is slower to search the first time a variable is referenced,
// that initial search will reveal the "lexical address" of the
// variable, which we can then store in place of the referring symbol.
// It is therefore necessary that variable storage in an environment
// never move, once allocated.  The simplest way to guarantee this is
// to manage the bindings ourselves in a vector; the lexical address
// can then be stored in the simple form of two integers and does not
// depend on peculiarities of the data-structure implementation.
//
// We overload the concept of Environment with other data needed to 
// evaluate Scheme expressions.  For example, Scheme I/O primitives
// like `with-input-from-file' provide for the presence of a stack
// of open files which we maintain in this structure.  

class Context
    {

    public:
    friend  class Cell;
    friend  class Slab;

    Context ();


    // "Binding" is the process of asserting a value for a 
    // variable in the given environment.  That is, we do 
    // not search upward in the enclosure chain for an 
    // existing binding; we create one in the current environement.
    // (The contrast is with `set', which does perform such 
    // a search.

    Cell *  extend (Cell * env);
    void    bind (Cell * env, Cell * c, Cell * value);
    void    bind (Cell * c, Cell * value)
	{ bind (envt, c, value); }
    void    bind_arguments (Cell * env, Cell * vars, Cell * values);

    // When new bindings are created, the existing environment
    // is _extended_ with a vector of new {variable,value} bindings
    // provided in parallel-list form.

    // Getting and Setting values in an environment is slightly
    // different from binding: `get' will search the enclosure
    // chain if necessary, returning the innermost matching binding.
    // Set does the same.  Both of these will signal an error if
    // a binding cannot be found (they will not establish one: only
    // bind can do that).

    Cell *	get (Cell * env, Cell * c);
    void 	set (Cell * env, Cell * var, Cell * value);

    // root : find the "root" (i.e., parentless) environment 
    // which contains this one.

    bool        read_eval_print (FILE * in, FILE * out, bool);
    Cell * 	eval (Cell * form);
    Cell *      root () { return root_envt; }
    void        gc ();
    void        gc_if_needed ();
    void        print_mem_stats (FILE *);
    
    // Convert text to live cells 

    Cell *      read (sio &);
    Cell *      read (FILE *);

    // Manufacture Cells and Atoms

    Cell * make ();
    Cell * make_int (int i);
    Cell * make_char (char ch);
    Cell * make_real (double d);
    Cell * make_string (const char * s);
    Cell * make_allocated_string (const char * s, int len);
    Cell * make_subr (subr_f s, const char * name);
    Cell * make_builtin (psymbol y);
    Cell * make_symbol (psymbol y);
    Cell * make_boolean (bool b);
    Cell * make_vector (int n, Cell * init = &Cell::Unspecified);
    Cell * make_iport (const char * fname);
    Cell * make_iport (FILE *);
    Cell * make_oport (const char * fname);
    Cell * make_oport (FILE * op);
    Cell * make_procedure (Cell * env, Cell * body, Cell * arglist);
    Cell * make_promise (Cell * env, Cell * body);
    Cell * make_macro (Cell * env, Cell * body, Cell * arglist);
    Cell * make_magic (void *, magic_set_f, magic_get_f);
    Cell * make (Cell * ca, Cell * cd = &Cell::Nil);
    Cell * make_list1 (Cell *);
    Cell * make_list2 (Cell *, Cell *);
    Cell * make_list3 (Cell *, Cell *, Cell *);
    Cell * make_continuation ();
    void   load_continuation (Cell * cont);

    Cell * cons (Cell * _car, Cell * _cdr) { return make (_car, _cdr); }

    // ------------------------------------------------------------

    void                with_input (const char * fname)
        {
        istack.push (make_iport (fname));
        }

    void                with_output (const char * fname)
        {
        ostack.push (make_oport (fname));
        }

    void                without_output ()
        {
	fflush (ostack.pop ()->OportValue ());
        }

    void                without_input ()
        {
        istack.pop ();
        }

    Cell *              current_output () {return ostack.top ();}
    Cell *              current_input  () {return istack.top ();}

    // Protection from garbage collection (cell pointers not contained
    // in "register machine" variables need to be treated this way.
    // The variables are protected/unprotected in strict LIFO order.

    void                gc_protect (Cell * c)
	{ r_gcp.push (c); }
    void                gc_unprotect (int ncells = 1)
	{ r_gcp.discard (ncells); }

    private:

    Cell *              alloc (Cell::Type t, int s = 1);
    void                mark (Cell *);
    Cell *              _eval (Cell * form);
    Cell *              find (Cell * env, Cell * c);
    void                quicken (Cell *, int, int);
    Cell * 		eval_list (Cell * list);
    void                provision ();
    void                init_machine ();
    void                print_vm_state ();

    // ===========================
    // Machine Stack Operations

    // The machine stack is just a cellvector, with one difference:
    // it can hold integers (marked with the ATOM flag) as well as 
    // cell pointers.  (There are thus only 31 bits in these integers,
    // but that's way more than enough to hold the virtual machine
    // state.

    void                save (Cell * c)      { m_stack.push (c);              }
    void                save (Cell & rc)     { m_stack.push (rc.ca.p);
                                               m_stack.push (rc.cd.p);        }
//    void                save (int i)
//	{ m_stack.push (reinterpret_cast <Cell *> ((i << 1) | Cell::ATOM)); }
    void save(int i) {m_stack.push(make_int(i));}
    void                restore (Cell *& c)  { c = m_stack.pop ();            } 
    void                restore (Cell & rc)  { rc.cd.p = m_stack.pop ();
                                               rc.ca.p = m_stack.pop ();      }
//    void                restore (int & i)
//	{ i = (reinterpret_cast <int> (m_stack.pop ()) & ~Cell::ATOM) >> 1; }
    void restore (int& i) 
	{ Cell* c = m_stack.pop(); i = c->IntValue(); } 

    // ===========================
    // REGISTER MACHINE
    // ===========================

    Cell *              r_exp;      // expression to evaluate
    Cell *              r_env;      // evaluation environment
    Cell *              r_unev;     // args awaiting evaluation
    Cell                r_argl;     // (head,tail) of argument list
    Cell                r_varl;     // (head,tail) of binding list
    Cell *              r_proc;     // procedure to apply
    Cell *              r_val;      // value resulting from evaluation
    Cell *              r_tmp;      // temporary values
    Cell *              r_elt;      // elements assembled into lists
    Cell *              r_nu;       // reference to objects being created
    int                 r_qq;       // quasiquotation depth
    cellvector          r_gcp;      // extra cells protected from GC
    int                 r_cont;     // current continuation
    cellvector          m_stack;    // recursion/evaluation stack
    int                 state;      // current machine state

    // ===========================
    
    // routines to append elements to lists (used with r_argl and r_varl).
    // Note: r_argl and r_varl MUST be maintained as correctly-formed
    // lists, since we use unsafe car/cdr to traverse them.  

    void l_appendtail (Cell & l, Cell * t)
	{
	if (l.ca.p == nil)
	    l.ca.p = l.cd.p = t;
	else
	    {
	    l.cd.p->cd.p = t; // l.cd.p->setcdr (t);
	    l.cd.p = t;
	    }
	}

    void l_append (Cell & l, Cell * t)
	{
	r_elt = make (t);
	l_appendtail (l, r_elt);
	}

    void clear (Cell & c) 
	{
	c.ca.p = c.cd.p = nil;
	}

    Cell *              envt;
    Cell *              root_envt;

    cellvector          istack;     // stack of input ports (with-input...)
    cellvector          ostack;     // stack of output ports (with-output...)

    struct              Memory
	{
	cellvector	active;         // list of allocated Slabs
	Cell *	        free1;	        // free single cells
	Cell *	        free2;	        // free DOUBLE cells
	int             c_free1;        // free ordinary cells
	int             c_free2;        // free double-deck cells
	Slab *          current ()      { return (Slab *) active.top (); } 
	bool            low_water;      // true if next exhaustion should alloc
	bool            last_alloc_gc;	// true if last allocation provoked gc
	bool            no_inline_gc;   // don't try gc on allocation failure

	Memory ()
	    : active ()
	    {
	    free1 = free2 = 0;
	    c_free1 = c_free2 = 0;
	    low_water = last_alloc_gc = no_inline_gc = false;
	    }
	};

    bool         ok_to_gc;
    Memory       mem;

    int          cellsAlloc;
    int          cellsTotal;
    };

class VxSchemeInit
    {
    public:
    
    VxSchemeInit ()
	{
	// Do sanity checks before scheme runs
	Cell::config ();
	}

    ~VxSchemeInit ()
	{
	// Print statistics when scheme exits.
	Cell::stats ();
	}
    };


// Simple accessors in prefix form `car (c)' rather than 
// C++ suffix form `c->car ()'.

inline Cell * car   (Cell * c)  {return c->car ();}
inline Cell * caar  (Cell * c)  {return c->caar ();}
inline Cell * cdr   (Cell * c)  {return c->cdr ();}
inline Cell * cdar  (Cell * c)  {return c->cdar ();}
inline Cell * cadr  (Cell * c)  {return c->cadr ();}
inline Cell * cddr  (Cell * c)  {return c->cddr ();}
inline Cell * cadar (Cell * c)  {return c->cadar ();}
inline Cell * caddr (Cell * c)  {return c->caddr ();}
inline Cell * caadr (Cell * c)  {return c->caadr ();}
inline Cell * cdadr (Cell * c)  {return c->cdadr ();}
inline Cell * cddar (Cell * c)  {return c->cddar ();}
inline Cell * caddar (Cell * c) {return c->caddar ();}
inline Cell * cadaar (Cell * c) {return c->cadaar ();}


// Certain syntactic features of Scheme (so-called "syntactic sugar"
// like the `else' clause in a cond statement, the use of `.' to
// construct improper lists and "varargs lambdas", and some of the
// mechanics of quasiquotation) are most easily implemented if we have
// predefined symbols for these tokens.  They are not part of the
// global environment, however, and have no definitions themselves.
// We create them with global scope (in the `C' sense) as they can
// serve as invariant hashcodes throughout any universe of Scheme
// execution: there is never any need to compute their values more
// than once, even for multiple threads.

extern psymbol s_dot;
extern psymbol s_quote;
extern psymbol s_quasiquote;
extern psymbol s_unquote;
extern psymbol s_unquote_splicing;
extern psymbol s_passto;
extern psymbol s_else;

// We treat special forms similarly.

extern psymbol s_if;
extern psymbol s_define;
extern psymbol s_quote;
extern psymbol s_begin;
extern psymbol s_set;
extern psymbol s_or;
extern psymbol s_and;
extern psymbol s_lambda;
extern psymbol s_let;
extern psymbol s_letstar;
extern psymbol s_letrec;
extern psymbol s_do;
extern psymbol s_cond;
extern psymbol s_case;
extern psymbol s_callcc;

// Cells can be inserted into streams in readable form.

// Execution flags

#define TRACE_EVAL		0x01
#define TRACE_GC		0x02
#define DEBUG_NO_INLINE_GC	0x04
#define DEBUG_MEMSTATS_AT_EXIT  0x08
#define DEBUG_PRINT_PROCEDURES  0x10
#define TRACE_GC_ALL            0x20
